;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((emacs-lisp-mode . ((fill-column . 90)
		     (indent-tabs-mode . nil)))
 (org-mode . ((org-list-two-spaces-after-bullet-regexp . nil))))
